/*=====================================================================*/

#include <stdio.h>
#include <assert.h>
#include "parti.h"

/*===================================================================*/

void  PARTI_fetch_lists( hash_table, mask, 
                                     fetch_p, fetch_l,  glb,  lcl,  fetch_size )
HASHSTR  *hash_table;
int        mask;
int       *fetch_p;              /* processor from which to fetch */
int       *fetch_l;              /* local offsets on those procs  */
int       *glb;
int       *lcl;
int       *fetch_size;           /* size of fetch_p, fetch_l      */
{

    int      no_buckets,  i,j, siz, k, max;
    int      start_offset;
    int      *key, *proc, *local,  *stamp, *offset;
    HASHCELL *buckets, *hashcell;
 

    assert( hash_table != NULL );

    no_buckets   = hash_table->no_buckets ;
    buckets      = hash_table->buckets ;
    siz = 0;

    for( i=0; i< no_buckets; i++ ) {
        for ( hashcell = &(buckets[i]) ;
              hashcell != NULL ; hashcell = hashcell->next ) {
            stamp  = hashcell->stamp ;
            offset = hashcell->offset ;
            proc   = hashcell->proc;
            local  = hashcell->local;
            key    = hashcell->key;
            for ( j=0; j < hashcell->tag; j++ ) {
               if ( mask & stamp[j] ) {
                    fetch_p[ siz ] = proc[j];
                    fetch_l[ siz ] = local[j];
                    lcl    [ siz ] = offset[j];
                    glb    [ siz ] = key[j];
                    siz ++ ;
               }
            }
        }
    }
    *fetch_size  = siz;
    return ;
}

/*===================================================================*/

void  PARTI_rep_fetch_lists(hash_table, mask, 
                                     fetch_p, fetch_l,  glb,  lcl,  fetch_size )
HASHSTR  *hash_table;
int        mask;
int       *fetch_p;              /* processor from which to fetch */
int       *fetch_l;              /* local offsets on those procs  */
int       *glb;
int       *lcl;
int       *fetch_size;           /* size of fetch_p, fetch_l      */
{

    int      no_buckets,  i,j, siz, k, max;
    int      start_offset;
    TTABLE   *trans_table;		/* translation table */
 

    assert( hash_table != NULL );

    trans_table = hash_table->trans_table;
    no_buckets   = hash_table->no_buckets ;
    siz = 0;

    for(i=0; i< no_buckets; i++) {
        if (mask & hash_table->cells[i].stamp) {
           fetch_p[siz] = trans_table->tabptr->Lproc[i];
           fetch_l[siz] = trans_table->tabptr->Loffset[i];
           lcl    [siz] = hash_table->cells[i].offset;
           glb    [siz] = i;
           siz ++ ;
        }
    }
    *fetch_size  = siz;
    return ;
}

/*======================================================================*/

void  PARTI_inc_fetch_lists( hash_table, mymask,
                                fetch_p, fetch_l,  glb,  lcl,  fetch_size )

HASHSTR   *hash_table;
int        mymask ;
int       *fetch_p;              /* processor from which to fetch */
int       *fetch_l;              /* local offsets on those procs  */
int       *glb;
int       *lcl;
int       *fetch_size;           /* size of fetch_p, fetch_l      */
{

    int      no_buckets,  i,j, siz, k, max;
    int      start_offset;
    int      mystamp ;
    int      *key, *proc, *local,  *stamp, *offset;
    int      check_mask;
    HASHCELL *bukets, *hashcell;
    int m; 
 

    assert( hash_table != NULL );

    no_buckets   = hash_table->no_buckets ;
    bukets       = hash_table->buckets ;
    check_mask   = incremental_mask_maker( mymask );  
    mystamp      = find_smallest_stamp( mymask ); 
    start_offset = hash_table->start_offset[mystamp];
    siz = 0;
    max = find_inc_size( hash_table, mymask);

    for( i=0; i<max; i++ ) fetch_p[i] = -1;


    for( i=0; i< no_buckets; i++ ) {
        for ( hashcell = &(bukets[i]) ;
              hashcell != NULL ; hashcell = hashcell->next ) {
            stamp  = hashcell->stamp ;
            offset = hashcell->offset ;
            proc   = hashcell->proc;
            local  = hashcell->local;
            key    = hashcell->key;
            for ( j=0; j < hashcell->tag; j++ ) {
               if ( ((mymask & stamp[j]) > 0) && 
		    ((check_mask & stamp[j])==0)) {
                    k   = offset[j]-start_offset;
		    if ( k >  max )  {
		      printf("\n offset %d ", offset[j]);
		    }
		    if ( fetch_p[k] != -1 ) {
                      m = glb[k];
                      printf("\n [%d] whats going on ?  fetch_p[%d] : %d, key : %d", 
			      PARTI_mynode, k, fetch_p[k], glb[k]);
                      printf("\n [%d],Offset[0] = %d , start_offset = %d, max = %d",
				  PARTI_mynode,offset[j] , start_offset, max);
                      exit(1);
		    }
                    fetch_p[ k ] = proc[j];
                    fetch_l[ k ] = local[j];
                    lcl    [ k ] = offset[j];
                    glb    [ k ] = key[j];
                    siz ++ ;
               }
            }
        }
    }
    *fetch_size  =  siz;
    return ;
}

/*======================================================================*/

void  PARTI_rep_inc_fetch_lists(hash_table, mymask,
                                fetch_p, fetch_l,  glb,  lcl,  fetch_size )

HASHSTR   *hash_table;
int        mymask ;
int       *fetch_p;              /* processor from which to fetch */
int       *fetch_l;              /* local offsets on those procs  */
int       *glb;
int       *lcl;
int       *fetch_size;           /* size of fetch_p, fetch_l      */
{

    int      no_buckets,  i,j, siz, k, max;
    int      start_offset;
    int      mystamp ;
    int      stamp;
    int      check_mask;
    int m; 
    TTABLE   *trans_table;		/* translation table */
 

    assert( hash_table != NULL );

    trans_table	 = hash_table->trans_table;
    no_buckets   = hash_table->no_buckets ;
    check_mask   = incremental_mask_maker( mymask );  
    mystamp      = find_smallest_stamp( mymask ); 
    start_offset = hash_table->start_offset[mystamp];
    siz = 0;
    max = find_inc_size( hash_table, mymask);

    for( i=0; i<max; i++ ) fetch_p[i] = -1;


    for( i=0; i< no_buckets; i++ ) {
	stamp = hash_table->cells[i].stamp;
	if ( ((mymask & stamp) > 0) && ((check_mask & stamp)==0)) {
            k = hash_table->cells[i].offset - start_offset;
	    if ( k > max )  {
		printf("\n offset %d ", hash_table->cells[i].offset);
	    }
	    if ( fetch_p[k] != -1 ) {
                m = glb[k];
                printf("\n [%d] whats going on ?  fetch_p[%d] : %d, key : %d", 
			      PARTI_mynode, k, fetch_p[k], glb[k]);
		printf("\n [%d],Offset[0] = %d , start_offset = %d, max = %d",
				  PARTI_mynode,hash_table->cells[i].offset, 
				  start_offset, max);
                exit(1);
	    }
            fetch_p[ k ] = trans_table->tabptr->Lproc[i];
            fetch_l[ k ] = trans_table->tabptr->Loffset[i];
            lcl    [ k ] = hash_table->cells[i].offset;
            glb    [ k ] = i;
            siz ++ ;
        }
    }
    *fetch_size  =  siz;
    return ;
}

/*===============================================================*/

int incremental_mask_maker( mask )
int  mask;
{

  int  i,j = 1;
  int  tmp = 0;
  int  btmp = 0;
  for ( i=0; i< MAX_STAMPS && j < mask ; j<<=1, i++) { 
    if ( (j & mask) == 0 )  tmp = tmp | j;
  }   
  return tmp ;
}

/*=================================================================*/

int find_smallest_stamp( mask )
int mask ;
{

int i    = 0;
int test = 1;

while (  !(mask & test ) ) {  mask >>= 1; i++; }
return i;
}
/*=======================================================================*/
